# Copyright (c) 2001-2005 Twisted Matrix Laboratories.
# Copyright (c) 2005-2006 James Bunton
# See LICENSE for details.


"""An epoll() based implementation of the twisted main loop.

To install the event loop (and you should do this before any connections,
listeners or connectors are added)::

    from twisted.internet import epollreactor
    epollreactor.install()

API Stability: stable

Maintainer: U{James Bunton<mailto:james@delx.cjb.net>}
"""

# System imports
import errno, sys

# Twisted imports
from twisted.epoll import epoll
from twisted.python import log, threadable, failure
from twisted.internet import main, posixbase, error

# globals
reads = {}
writes = {}
masks = {}
selectables = {} #fd->selectable
fds = {} #selectable->fd
poller = epoll.ePoller()

POLL_READ = epoll.EPOLLIN | epoll.EPOLLPRI
POLL_WRITE = epoll.EPOLLOUT
POLL_DISCONNECTED = (epoll.EPOLLHUP | epoll.EPOLLERR)

class ePollReactor(posixbase.PosixReactorBase):
    """A reactor that uses epoll(2)."""
    def __init__(self):
        posixbase.PosixReactorBase.__init__(self)


    def _selectableRemove(self, selectable, fd):
        if not selectables.has_key(fd) and fds.has_key(selectable):
            fd = fds[selectable]
        if selectables.has_key(fd):
            del selectables[fd]

        if not fds.has_key(selectable) and selectables.has_key(fd):
            selectable = selectables[fd]
        if fds.has_key(selectable):
            del fds[selectable]


    def _tryModregister(self, fd, mask):
        try:
            poller.modregister(fd, mask)
        except epoll.ePollError, e:
            if e.value == errno.ENOENT:
                poller.register(fd, mask)
            else:
                raise

    def _tryUnregister(self, fd):
        try:
            poller.unregister(fd)
        except epoll.ePollError, e:
            if e.value != errno.ENOENT:
                raise

    def addReader(self, reader):
        """Add a FileDescriptor for notification of data available to read.
        """
        fd = reader.fileno()
        if not reads.has_key(fd):
            selectables[fd] = reader
            fds[reader] = fd
            reads[fd] =  1
            mask = POLL_READ
            if writes.has_key(fd):
                mask |= POLL_WRITE
                self._tryModregister(fd, mask)
            else:
                poller.register(fd, mask)

    def addWriter(self, writer, writes=writes, selectables=selectables):
        """Add a FileDescriptor for notification of data available to write.
        """
        fd = writer.fileno()
        if not writes.has_key(fd):
            selectables[fd] = writer
            fds[writer] = fd
            writes[fd] =  1
            mask = POLL_WRITE
            if reads.has_key(fd):
                mask |= POLL_READ
                self._tryModregister(fd, mask)
            else:
                poller.register(fd, mask)

    def removeReader(self, reader, reads=reads):
        """Remove a Selectable for notification of data available to read.
        """
        if not fds.has_key(reader): return
        fd = fds[reader]

        if reads.has_key(fd): del reads[fd]
        try:
            if writes.has_key(fd):
                self._tryModregister(fd, POLL_WRITE)
            else:
                self._selectableRemove(reader, fd)
                self._tryUnregister(fd)
        except epoll.ePollError, e:
            if e.value == errno.EBADF:
                pass
            else:
                raise

    def removeWriter(self, writer, writes=writes):
        """Remove a Selectable for notification of data available to write.
        """
        if not fds.has_key(writer): return
        fd = fds[writer]

        if writes.has_key(fd): del writes[fd]
        try:
            if reads.has_key(fd):
                self._tryModregister(fd, POLL_READ)
            else:
                self._selectableRemove(writer, fd)
                self._tryUnregister(fd)
        except epoll.ePollError, e:
            if e.value == errno.EBADF:
                pass
            else:
                raise

    def removeAll(self, reads=reads, writes=writes, selectables=selectables):
        """Remove all selectables, and return a list of them."""
        if self.waker is not None:
            self.removeReader(self.waker)
        result = selectables.values()
        selectables.clear()
        for fd in fds.values():
            self._tryUnregister(fd)
        fds.clear()
        reads.clear()
        writes.clear()

        if self.waker is not None:
            self.addReader(self.waker)
        return result

    def doPoll(self, timeout,
               reads=reads,
               writes=writes,
               selectables=selectables,
               log=log):
        """Poll the poller for new events."""
        if timeout is None:
            timeout = 0
        else:
            timeout = int(timeout * 1000) # convert seconds to milliseconds

        try:
            l = poller.poll(timeout)
        except epoll.ePollError, e:
            if e.value == errno.EINTR:
                return
            else:
                raise
        _drdw = self._doReadOrWrite
        for fd, event in l:
            try:
                selectable = selectables[fd]
            except KeyError:
                # Handles the infrequent case where one selectable's
                # handler disconnects another.
                continue
            log.callWithLogger(selectable, _drdw, selectable, fd, event, log)

    doIteration = doPoll

    def _doReadOrWrite(self, selectable, fd, event, log, 
        faildict={
            error.ConnectionDone: failure.Failure(error.ConnectionDone()),
            error.ConnectionLost: failure.Failure(error.ConnectionLost())
        }):
        why = None
        inRead = False
        if event & POLL_DISCONNECTED and not (event & (epoll.EPOLLIN | epoll.EPOLLPRI)):
            why = main.CONNECTION_LOST
        else:
            try:
                if (event & (epoll.EPOLLIN | epoll.EPOLLPRI)):
                    why = selectable.doRead()
                    inRead = True
                if not why and event & epoll.EPOLLOUT:
                    why = selectable.doWrite()
                    inRead = False
                if not selectable.fileno() == fd:
                    why = error.ConnectionFdescWentAway('Filedescriptor went away')
                    inRead = False
            except:
                log.deferr()
                why = sys.exc_info()[1]
        if why:
            self._disconnectSelectable(selectable, why, inRead)


def install():
    """Install the epoll() reactor."""
    p = ePollReactor()
    import main
    main.installReactor(p)
    return p

__all__ = ["ePollReactor", "install"]

